/* compiler.cpp
 *
 * Date: 3-07-2006
 * Author: Nicolas Pope
 *
 * Description: The interpreter interface to the core database. Provides methods
 * for converting objects to standard c++ data types and a means of converting OIDs
 * into strings.
*/

#include <Compiler.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <Index.h>
#include <Conversion.h>

using namespace doste;
using namespace std;

OID Compiler::m_current;
CompilerNotifiable *Compiler::notif = 0;

/* Called straight after Core::Init() to set up the indexes. The indexes are
 * loaded from the database and added to the cache hash table. A notification
 * is also added to the index object to respond to changes.*/
void Compiler::Init()
{
	m_current = Null;

	Index::Add(Null, "null");
	Index::Add(This, "this");
	Index::Add(Size, "size");
	Index::Add(System, "system");
	Index::Add(True, "true");
	Index::Add(False, "false");
	Index::Add(Parent, "..");
	Index::Add(Star, "*");
	Index::Add(Infinity, "infinity");
	Index::Add(Current, "current");
}

/* Called just before Core::Final to save the last used index number and
 * destroy the local index cache.*/
void Compiler::Final()
{
	
}

void Compiler::SetCurrent(OID cur)
{
	m_current = cur;
}

OID Compiler::GetCurrent()
{
	return m_current;
}

/* A very important part of the interpreter, it extracts individual tokens from
 * the input buffer stored in Session 's'. If it is in peek mode then it does
 * not move on to the next token. It also automatically removes white space and
 * comments.*/
Token Compiler::GetToken(Session *s, bool peek)
{
	//Remove all white space from the buffer, reguardless of whether its in peek mode.
	while (s->curbuf[0] == ' ' || s->curbuf[0] == '\n' || s->curbuf[0] == '\r' || s->curbuf[0] == '\t')
	{
		if (s->curbuf[0] == '\n')
			s->line++;
		s->curbuf++;
	}

	switch (s->curbuf[0])
	{
		case COMMENT:	s->curbuf++;
						while (s->curbuf[0] != '\n' && s->curbuf[0] != 0)			//Remove comment until new line
							s->curbuf++;
						return GetToken(s, peek);				//Call self to find next real token
		case CMD:		if (!peek) s->curbuf++; return CMD;		//If '%' then load file
		case TERM:		return TERM;							//End of buffer, null.
		case TLP:		if (!peek) s->curbuf++; return TLP;		//A '(' character
		case RP:		if (!peek) s->curbuf++; return RP;		//')'
		case LS:		if (!peek) s->curbuf++; return LS;		//'{' for sets
		case RS:		if (!peek) s->curbuf++; return RS;		//'}'
		case LO:		if (!peek) s->curbuf++; return LO;		//'<', unused, was for definitions
		case RO:		if (!peek) s->curbuf++; return RO;		//'>'
		case ASSIGN:	if (!peek) s->curbuf++; return ASSIGN;	//'='
		case SEP:		if (!peek) s->curbuf++; return SEP;		//',' for use in sets
		case END:		if (!peek) s->curbuf++; return END;		//';' to terminate a statement

		default:		s->tokval = TranslateObject(s, peek);	//Otherwise its an object name or id
						return OBJECT;							//Interpret name/id and return OBJECT token
	}
}

/* Given a string version of an OID, eg. "<1:167>", it converts it into
 * a real OID. This is an unsafe implementation in that it writes to the
 * passed string parameter which may be read only!*/
OID Compiler::TranslateOID(char *word)
{
	word++;										//Remove '<'
	char *buf = word;
	while (buf[0] != (char)OIDSEP && buf[0] != 0)		//Find the ':' and stop
		buf++;
	buf[0] = 0;									//Make into temporary string end
	buf++;										//Move buffer to part after ':'
	OID result;
	if (word[0] == 0)							//Allow shorthand "<:167>" to mean "<1:167>"
		result.type = OID_TYPE_BASE;
	else
		result.type = atoi(word);				//Convert part before ':' to an integer type id
	word = buf;									//Move on to second part
	while (buf[0] != (char)RC && buf[0] != 0)			//Find '>' and stop
		buf++;
	buf[0] = 0;									//Set as temporary end of string
	result.value = atoi(word);						//Convert second part to an integer.

	return result;
}

/* Reads an object name, id, number or string from the input buffer and
 * translates into an OID.*/
OID Compiler::TranslateObject(Session *s, bool peek)
{
	const char *cur = s->curbuf;
	OID result;
	char word[200];							//Store words in new buffer
	int pos = 0;
	bool str = false;						//Not unside a string
	bool chr = false;						//Not inside a character

	//If in a string or character or not a deliminating character then
	while (chr || str || (cur[0] != (char)LO && cur[0] != (char)RO && cur[0] != (char)TLP && cur[0] != (char)RP && cur[0] != (char)LS && cur[0] != (char)RS && cur[0] != (char)ASSIGN && cur[0] != (char)SEP
			&& cur[0] != (char)END && cur[0] != (char)TERM
			&& cur[0] != ' ' && cur[0] != '\t' && cur[0] != '\n' && cur[0] != '\r'))
	{
		if (cur[0] == '\n')
			s->line++;

		if (!chr && cur[0] == '"')			//If not inside a character then
			str = !str;						//Either become a string or end a string
		else if	(!str && cur[0] == '\'')	//If not inside a string then
			chr = !chr;						//Either become a character or end a character
		word[pos] = cur[0];					//Add character to the word buffer
		cur++;								//Move to next input buffer location
		pos++;								//Move to next word buffer location
	}
	word[pos] = 0;							//Add terminating null to string

	switch (word[0])						//Check first character to get type
	{
		case '-':
		case '0':							//If any numerical digit
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':	if (strchr(word, '.') == 0)			//Then check for a '.' for floating point
					{
						result.type = OID_TYPE_INTEGER;		//If not then its an integer
						result.value = atoi(word);			//So convert it to an integer
					} else {
						result = FloatToObj(atof(word));	//Otherwise it is so convert it to a float
					}
					break;

		case '\'':	result.type = OID_TYPE_CHAR;		//Its a character
					result.value = word[1];					//Use next character as character
					break;								//Should allow for escape characters!!!

		case '"':	word[pos-1] = 0;					//Its a string
					if (!peek)
						result = StrToObj(&word[1]);	//So create a new string object for it
					else
						result = Null;			//If peeking then don't create string object
					break;

		case LC:	result = TranslateOID(word); break;	//Its an explicit OID so convert

		default:	result = Index::Lookup(word);			//Otherwise it's an index label so look it up.
				//if (result == This)
				//	result = s->thi;
				if (result == Current)
					result = Object::GetCurrent();
	}

	if (!peek)					//Update input buffer to end if not peeking
		s->curbuf = cur;

	return result;
}

OID Compiler::EvalDef(vector<OID> *def)
{
	if (def == 0)
		return Null;

	OID res = (*def)[0];
	Object *obj;
	Object *lastobj = 0;
	OID temp;


	if (def->size() == 1)
		return res;

	for (unsigned int i=1; i<def->size(); i++)
	{
		obj = Object::Lookup(res);
		if (obj != 0)
		{
			obj->Open();

			if (lastobj != 0)
				lastobj->Close();

			temp = (*def)[i];
			if (temp.type == OID_TYPE_DEF)
				temp = EvalDef((vector<OID>*)(temp.value));
			res = obj->Get(temp);

			lastobj = obj;
			//obj->Close();
		} else
		{
			res = Null;
			break;
		}
	}

	if (lastobj != 0)
		lastobj->Close();

	return res;
}

void Compiler::PrintError(Session *s, const QString &error)
{
	if (notif != 0)
	{
		if (s != 0)
		{
			QString err = QString().sprintf("%s:%i: error: ", s->filename, s->line);
			err += error;
			notif->PrintError(err);
		} else
		{
			QString err = QString("error: ");
			err += error;
			notif->PrintError(err);
		}
	}
}

void Compiler::PrintWarning(Session *s, const QString &warning)
{
	if (notif != 0)
	{
		if (s != 0)
		{
			QString err = QString().sprintf("%s:%i: warning: ", s->filename, s->line);
			err += warning;
			notif->PrintWarning(err);
		} else
		{
			QString err = QString("error: ");
			err += warning;
			notif->PrintWarning(err);
		}
	}
}

void Compiler::SetNotifiable(CompilerNotifiable *notif)
{
	Compiler::notif = notif;
}

int Compiler::ParseCommand(Session *s)
{
	char buf[20];
	int pos = 0;

	while (pos < 19 && s->curbuf[0] != ' ' && s->curbuf[0] != '\n' && s->curbuf[0] != '\t' && s->curbuf[0] != 0)
	{
		buf[pos] = s->curbuf[0];
		s->curbuf++;
		pos++;
	}

	buf[pos] = 0;

	if (strcmp(buf, "include") == 0)
		return CMD_INCLUDE;
	if (strcmp(buf, "list") == 0)
		return CMD_LIST;
	if (strcmp(buf, "def") == 0)
		return CMD_DEF;
	if (strcmp(buf, "using") == 0)
		return CMD_USING;
	if (strcmp(buf, "query") == 0)
		return CMD_QUERY;
	if (strcmp(buf, "debug") == 0)
		return CMD_DEBUG;
	if (strcmp(buf, "print") == 0)
		return CMD_PRINT;
	if (strcmp(buf, "hide") == 0)
		return CMD_HIDE;

	s->errstatement = true;
	PrintError(s, QString().sprintf("Unknown command %%%s.", buf));
	return CMD_UNKNOWN;
}

void Compiler::ParseFilename(Session *s, char *buffer)
{
	int pos = 0;
	while (s->curbuf[0] == ' ' || s->curbuf[0] == '\n' || s->curbuf[0] == '\r' || s->curbuf[0] == '\t')
		s->curbuf++;

	if (s->curbuf[0] == '"')				//Filename must be in '"'.
	{
		s->curbuf++;						//Remove '"'
		while (s->curbuf[0] != '"' && s->curbuf[0] != 0 && s->curbuf[0] != ';' && s->curbuf[0] != '\n')			//Until end of filename string
		{
			buffer[pos] = s->curbuf[0];		//Add character to filename buffer
			s->curbuf++;
			pos++;
		}

		if (s->curbuf[0] != '"')
		{
			PrintError(s, QString("No terminating '\"' after file name."));
			s->errstatement = true;
			return;
		}

		s->curbuf++;

		//The remove any extra characters from buffer until either ';' or null.
		while(s->curbuf[0] != (char)TERM && s->curbuf[0] != (char)END)
		{
			//if (s->curbuf[0] != ' ' && s->curbuf[0] != '\n' && s->curbuf[0] != '\r' && s->curbuf[0] != '\t')
				//MessageBox(0, "Possibly missing a ';'", "Interpreter Error", MB_OK | MB_ICONWARNING);
			s->curbuf++;
		}

		buffer[pos] = 0;					//Add terminating null to end of file buffer
		return;			//Execute the file and return result.
	} else {
		PrintError(s, QString("Missing '\"' around file name."));
		s->errstatement = true;
		return;
	}
}

/* Parses a single statement up to a ';' from the input. A statement can be one of 4 things...
 * A straight forward get request, and assignment, a load file request or (not yet done) a
 * remove attribute request.*/
OID Compiler::ParseStatement(Session *s)
{
	Token tok;
	OID res, key, temp, temp2;
	Object *object;
	vector<OID> *res2 = 0;
	vector<OID> *lhsdef = 0;
	//Attribute *attrib;
	int i;
	bool getdef = s->getdef;
	s->getdef = false;

	//s->thi = Null;
	s->lhs = true;								//REDUNDANT?
	tok = GetToken(s, true);					//Peek at first token
	
	if (tok == CMD)								//If its a '%' then execute a file
	{
		char buffer[200];						//Buffer for file name
		GetToken(s, false);						//Remove token

		int cmd = ParseCommand(s);

		if (s->critical || s->errstatement)
			return Null;

		switch(cmd)
		{
			case CMD_INCLUDE:		ParseFilename(s, buffer);
									if (s->errstatement || s->critical)
										return Null;
									return ExecuteFile(buffer);

			case CMD_LIST:			temp = ParseStatement(s);
									object = Object::Lookup(temp);
									if (object != 0)
									{
										object->Open();
										i = 0;
										temp = object->GetAttribute(i);
										while (temp != Error)
										{
											OIDToString(temp, buffer, 48);
											if (notif) notif->PrintText(QString().sprintf("&nbsp;&nbsp;%s", buffer));
											i++;
											temp = object->GetAttribute(i);
										}
										object->Close();
									}
									return Null;

			case CMD_DEF:			if (notif != 0) notif->PrintText(QString("DEF")); break;

			case CMD_USING:			Object::SetCurrent(ParseStatement(s));
									break;

			case CMD_HIDE:			s->getdef = true;
									ParseStatement(s);
									if (s->errstatement)
									{
										return Null;
									}
									temp2 = s->def->back();
									s->def->pop_back();
									temp = EvalDef(s->def);
									object = Object::Lookup(temp);
									delete s->def;
									if (object != 0)
									{
										object->Open();
										object->SetFlags(temp2, (object->GetFlags(temp2) | FLAG_HIDDEN));
										object->Close();
									} else
									{
										PrintError(s, QString("No object to hide."));
										s->errstatement = true;
										return Null;
									}
									break;

			case CMD_QUERY:			s->query = true;
									temp = ParseStatement(s);
									s->query = false;
									return temp;

			case CMD_DEBUG:			s->debug = true;
									//Object::info = 0xFFFF;
									temp = ParseStatement(s);
									//Object::info = 0;
									s->debug = false;
									return temp;

			case CMD_PRINT:			ParseFilename(s, buffer);
									PrintWarning(s, buffer);
									return Null;

			default: break;
		}

		return Null;

	} else if (tok != TERM && tok != END)		//If not '%' then continue as normal
	{
		lhsdef = ParseObject(s, Null);
		if (getdef)
		{
			s->def = lhsdef;
			res = Null;
		} else
			res = EvalDef(lhsdef);		//Parse LHS as an object.
		if (s->critical || s->errstatement)
		{
			delete lhsdef;
			return Null;
		}
	}
	else
	{
		PrintWarning(s, QString("No Statement."));
		return Null;
	}

	tok = GetToken(s, true);					//Peek at next token
	if (tok == ASSIGN)							//If assignment then
	{
		if (res == Null)				//Make sure LHS was valid
		{
			char buf1[50];
			OIDToString(s->key, buf1, 48);
			PrintError(s, QString().sprintf("LHS of assignment is 'null', the key is '%s.'", buf1));
			s->errstatement = true;
			if (lhsdef != 0)
				delete lhsdef;
			return Null;
		}

		GetToken(s, false);						//Remove '='
		tok = GetToken(s, true);				//Check the next token to make sure its a valid assignment
		if (tok == ASSIGN || tok == END || tok == TERM)
		{
			PrintError(s, QString("No definition on RHS of assignment."));
			s->errstatement = true;
			if (lhsdef != 0)
				delete lhsdef;
			return Null;
		}

		s->lhs = false;							//REDUNDANT?
		key = s->key;							//Get last key from LHS
		s->key = Error;					//Reset the key in the Session object
		s->thi = res;
		res2 = ParseObject(s, Null);	//Parse RHS object

		if (s->critical || s->errstatement)
		{
			if (lhsdef != 0)
				delete lhsdef;
			if (res2 != 0)
				delete res2;
			return Null;
		}

		object = Object::Lookup(res);			//Open LHS object
		if (object != 0)
		{
			object->Open();
			if (res2->size() == 1)
			{
				object->Set(key, (*res2)[0]);
				delete res2;
			} else
				object->Set(key, res2);
			res2 = 0;
			object->Close();
		} else
		{
			PrintError(s, QString("Object on LHS does not exist."));
			s->errstatement = true;
			if (lhsdef != 0)
				delete lhsdef;
			if (res2 != 0)
				delete res2;
			return res;
		}
		tok = GetToken(s, true);				//Peek at next token

		if (s->query && notif != 0)
		{
			char buf1[50];
			char buf2[50];
			char buf3[50];
			OIDToString(res, buf1, 48);
			OIDToString(key, buf2, 48);
			object = Object::Lookup(res);
			if (object != 0)
			{
				object->Open();
				res = object->Get(key);
				object->Close();
			} else
				res = Error;
			OIDToString(res, buf3, 48);
			notif->PrintText(QString().sprintf("%s(%s) = %s", buf1, buf2, buf3));
		}

		res = Null;
	} else
	{
		if (notif != 0 && s->query)
		{
			char buf1[50];
			char buf2[50];
			char buf3[50];
			OIDToString((*lhsdef)[lhsdef->size()-1], buf2, 48);
			lhsdef->pop_back();
			temp = EvalDef(lhsdef);
			OIDToString(temp, buf1, 48);
			OIDToString(res, buf3, 48);
			if (temp != res)
				notif->PrintText(QString().sprintf("%s(%s) = %s", buf1, buf2, buf3));
			else
				notif->PrintText(buf1);
		}
	}

	if (tok != END && tok != TERM)
	{
		PrintWarning(s, QString("Missing ';' at end of statement."));
		if (lhsdef != 0)
				delete lhsdef;
		if (res2 != 0)
			delete res2;
		return res;
	}

	//if (tok == END)
	//	GetToken(s, false);

	if (lhsdef != 0 && getdef == false)
		delete lhsdef;
	if (res2 != 0)
		delete res2;
	return res;
}


vector<OID> *Compiler::ParseObject(Session *s, OID self)
{
	Token tok;
	tok = GetToken(s, true);
	vector<OID> *res = new vector<OID>();
	//Object *object;
	//Attribute *attrib;
	vector<OID> *tdef;
	//s->key = Null;
	bool ceval = false;
	OID cevalres;

	if (tok == LS)
	{
		//GetToken(s, false);
		res->push_back(ParseSet(s));
		if (s->critical || s->errstatement)
		{
			delete res;
			return 0;
		}
	} else if (tok == OBJECT)
	{
		GetToken(s, false);
		res->push_back(s->tokval);
	} else
	{
		s->errstatement = true;
		PrintError(s, QString("Expecting a '{' or an object identifier."));
		delete res;
		return 0;
	}

	tok = GetToken(s, true);

	while (tok == TLP || tok == LO)
	{
		GetToken(s, false);

		if (s->key != Error)
		{
			res->push_back(s->key);
			s->key = Error;
			if (ceval)
			{
				cevalres = EvalDef(res);
				res->clear();
				res->push_back(cevalres);
			}
		}

		if (tok == LO)
		{
			if (res->size() > 1)
			{
				PrintError(s, QString("Cannot use '<' after a '('."));
				s->errstatement = true;
				return res;
			} else
				ceval = true;
		} else
			ceval = false;

		tdef = ParseObject(s, self);

		if (s->critical || s->errstatement)
			return res;

		if (tdef->size() == 1)
		{
			s->key = (*tdef)[0];
			delete tdef;
		} else
		{
			s->key.type = OID_TYPE_DEF;
			s->key.value = (unsigned int)tdef;
		}

		tok = GetToken(s, false);
		if ((tok != RP && ceval == false) || (tok != RO && ceval))
		{
			PrintError(s, QString("Missing ')' or '>' after an index object."));
			s->errstatement = true;
			return res;
		}

		tok = GetToken(s, true);
	}

	if ((tok == END || tok == TERM || tok == RP || tok == RC || tok == SEP || tok == RO))
	{
		if (s->key != Error)
		{
			res->push_back(s->key);
			s->key = Error;
			if (ceval)
			{
				cevalres = EvalDef(res);
				res->clear();
				res->push_back(cevalres);
			}
		}
		return res;
	} else if (tok == ASSIGN)
	{
		return res;
	} else
	{
		s->errstatement = true;
		PrintError(s, QString().sprintf("Unexpected symbol '%c'.", tok));
		return res;
	}
}

OID Compiler::ParseSet(Session *s)
{
	Token tok;
	OID res;
	vector<OID> *key, *value;
	Object *object;
	OID oldthi;

	tok = GetToken(s, false);
	if (tok != LS)
	{
		return Null;
	}

	object = new Object();
	object->Open();
	object->Set(Parent, s->thi);
	res = object->GetOID();

	oldthi = s->thi;
	s->thi = res;

	s->inset++;

	tok = GetToken(s, true);
	while (tok != RS)
	{
		GetToken(s, false);
		if (tok != TLP)
		{
			s->errstatement = true;
			PrintError(s, QString("Expecting a '(' for a key,value pair."));
			s->thi = oldthi;
			object->Close(false);
			return res;
		}
		key = ParseObject(s, res);
		if (s->critical || s->errstatement)
		{
			delete key;
			s->thi = oldthi;
			object->Close(false);
			return res;
		}
		tok = GetToken(s, false);
		if (tok != SEP)
		{
			delete key;
			s->errstatement = true;
			PrintError(s, QString("Invalid key,value pair, expecting ','."));
			s->thi = oldthi;
			object->Close(false);
			return res;
		}
		value = ParseObject(s, res);
		if (s->critical || s->errstatement)
		{
			delete key;
			delete value;
			s->thi = oldthi;
			object->Close(false);
			return res;
		}

		if (value->size() == 1)
		{
			if (key->size() == 1)
			{
				object->Set((*key)[0], (*value)[0]);
				delete key;
			} else
			{
				object->Set(key, (*value)[0]);
			}
			delete value;
		} else
		{
			if (key->size() == 1)
			{
				object->Set((*key)[0], value);
				delete key;
			} else
			{
				object->Set(key, value);
			}
		}

		tok = GetToken(s, false);
		if (tok != RP)
		{
			s->errstatement = true;
			PrintError(s, QString("Missing ')' after a key,value pair."));
			s->thi = oldthi;
			object->Close(false);
			return res;
		}
		tok = GetToken(s, true);
		if (tok == SEP)
		{
			GetToken(s, false);
			tok = GetToken(s,true);
		}
	}

	s->inset--;

	GetToken(s, false);
	s->thi = oldthi;
	object->Close(false);
	return res;
}

OID Compiler::ExecuteFile(const char *filename)
{
	FILE *file = fopen(filename, "r+b");
	OID res;

	if (file == 0)
	{
		PrintError(0, QString().sprintf("File '%s' not found.", filename));
		return Error;
	}

	int size;
	fseek(file, 0, SEEK_END);
	size = ftell(file);
	fseek(file, 0, SEEK_SET);

	if (size > 10000000)
	{
		fclose(file);
		PrintError(0, QString().sprintf("File '%s' was too large", filename));
		return Error;
	}

	char *buffer = new char[size+1];

	int bread = fread(buffer, 1, size, file);
	buffer[size] = 0;

	if (bread != size)
	{
		delete buffer;
		fclose(file);
		PrintError(0, QString().sprintf("Could not read '%s'.", filename));
		return Error;
	}

	res = Execute(buffer, filename);

	fclose(file);
	delete [] buffer;

	return res;
}

OID Compiler::Execute(const char *buffer, const char *filename)
{
	Session *s = new Session;
	s->buffer = buffer;
	s->curbuf = buffer;
	s->lhs = true;
	s->thi = Null;
	s->key = Error;
	s->inset = 0;
	s->line = 1;
	s->filename = filename;
	s->critical = false;
	s->errstatement = false;
	s->query = false;
	s->debug = false;
	s->getdef = false;

	Token tok = END;
	OID result;
	int epos = 0;

	while (tok != TERM)
	{
		epos = 0;
		result = ParseStatement(s);

		if (s->errstatement)
		{
			//Remove rest of statement.
			while ((s->curbuf[0] != (char)TERM) && (s->curbuf[0] != (char)END))
			{
				if (s->curbuf[0] == '\n')
					s->line++;
				s->curbuf++;
			}

			s->errstatement = false;
			s->key = Null;
			s->lhs = true;
			s->inset = 0;
			s->thi = Null;
		}

		if (s->critical)
			break;

		tok = GetToken(s, false);
		if (tok == END)
			tok = GetToken(s, true);
	}
	
	delete s;
	return result;
}

vector<OID> *Compiler::CompileDefinition(const char *buffer, OID parent)
{
	Session *s = new Session;
	s->buffer = buffer;
	s->curbuf = buffer;
	s->lhs = true;
	s->thi = parent;
	s->key = Error;
	s->inset = 0;
	s->line = 1;
	s->filename = "";
	s->critical = false;
	s->errstatement = false;
	s->query = false;
	s->debug = false;
	s->getdef = true;

	Token tok = END;
	vector<OID> *result;
	int epos = 0;

	while (tok != TERM)
	{
		epos = 0;
		ParseStatement(s);
		result = s->def;

		if (s->errstatement)
		{
			//Remove rest of statement.
			while ((s->curbuf[0] != (char)TERM) && (s->curbuf[0] != (char)END))
			{
				if (s->curbuf[0] == '\n')
					s->line++;
				s->curbuf++;
			}

			s->errstatement = false;
			s->key = Null;
			s->lhs = true;
			s->inset = 0;
			s->thi = Null;
		}

		if (s->critical)
			break;

		tok = GetToken(s, false);
		if (tok == END)
			tok = GetToken(s, true);
	}
	
	delete s;
	return result;
}
